﻿'This library includes the common error-correction used in NAND memory.
'A commercial license is required to use this code

Namespace ECC_LIB

    Public Interface EccEngine
        Function GenerateECC(block() As Byte) As Byte()
        Function ProcessECC(ByRef block() As Byte, ByRef stored_ecc() As Byte) As ECC_DECODE_RESULT
        Function ECC_Size() As Integer
        Sub Close()
    End Interface

    '512 byte processing (24-bit ECC) 1-bit correctable/2-bit detection
    Public Class Hamming : Implements EccEngine

        Sub New()

        End Sub

        Private byte_parity_table() As Byte = {
                &HFF, &HD4, &HD2, &HF9, &HCC, &HE7, &HE1, &HCA,
                &HCA, &HE1, &HE7, &HCC, &HF9, &HD2, &HD4, &HFF,
                &HB4, &H9F, &H99, &HB2, &H87, &HAC, &HAA, &H81,
                &H81, &HAA, &HAC, &H87, &HB2, &H99, &H9F, &HB4,
                &HB2, &H99, &H9F, &HB4, &H81, &HAA, &HAC, &H87,
                &H87, &HAC, &HAA, &H81, &HB4, &H9F, &H99, &HB2,
                &HF9, &HD2, &HD4, &HFF, &HCA, &HE1, &HE7, &HCC,
                &HCC, &HE7, &HE1, &HCA, &HFF, &HD4, &HD2, &HF9,
                &HAC, &H87, &H81, &HAA, &H9F, &HB4, &HB2, &H99,
                &H99, &HB2, &HB4, &H9F, &HAA, &H81, &H87, &HAC,
                &HE7, &HCC, &HCA, &HE1, &HD4, &HFF, &HF9, &HD2,
                &HD2, &HF9, &HFF, &HD4, &HE1, &HCA, &HCC, &HE7,
                &HE1, &HCA, &HCC, &HE7, &HD2, &HF9, &HFF, &HD4,
                &HD4, &HFF, &HF9, &HD2, &HE7, &HCC, &HCA, &HE1,
                &HAA, &H81, &H87, &HAC, &H99, &HB2, &HB4, &H9F,
                &H9F, &HB4, &HB2, &H99, &HAC, &H87, &H81, &HAA,
                &HAA, &H81, &H87, &HAC, &H99, &HB2, &HB4, &H9F,
                &H9F, &HB4, &HB2, &H99, &HAC, &H87, &H81, &HAA,
                &HE1, &HCA, &HCC, &HE7, &HD2, &HF9, &HFF, &HD4,
                &HD4, &HFF, &HF9, &HD2, &HE7, &HCC, &HCA, &HE1,
                &HE7, &HCC, &HCA, &HE1, &HD4, &HFF, &HF9, &HD2,
                &HD2, &HF9, &HFF, &HD4, &HE1, &HCA, &HCC, &HE7,
                &HAC, &H87, &H81, &HAA, &H9F, &HB4, &HB2, &H99,
                &H99, &HB2, &HB4, &H9F, &HAA, &H81, &H87, &HAC,
                &HF9, &HD2, &HD4, &HFF, &HCA, &HE1, &HE7, &HCC,
                &HCC, &HE7, &HE1, &HCA, &HFF, &HD4, &HD2, &HF9,
                &HB2, &H99, &H9F, &HB4, &H81, &HAA, &HAC, &H87,
                &H87, &HAC, &HAA, &H81, &HB4, &H9F, &H99, &HB2,
                &HB4, &H9F, &H99, &HB2, &H87, &HAC, &HAA, &H81,
                &H81, &HAA, &HAC, &H87, &HB2, &H99, &H9F, &HB4,
                &HFF, &HD4, &HD2, &HF9, &HCC, &HE7, &HE1, &HCA,
                &HCA, &HE1, &HE7, &HCC, &HF9, &HD2, &HD4, &HFF}

        Private Function BitCount(data() As Byte) As Integer
            Dim counter As Integer
            For i = 0 To data.Length - 1
                Dim temp As Byte = data(i)
                Do While (temp > 0)
                    If (temp And 1) = 1 Then counter += 1
                    temp = (temp >> 1)
                Loop
            Next
            Return counter
        End Function

        Public Function GenerateECC(block() As Byte) As Byte() Implements EccEngine.GenerateECC
            Dim ecc_code(2) As Byte
            Dim word_reg As UInt16
            Dim LP0, LP1, LP2, LP3, LP4, LP5, LP6, LP7, LP8, LP9, LP10, LP11, LP12, LP13, LP14, LP15, LP16, LP17 As UInt32
            Dim uddata() As UInt32 = Utilities.Bytes.ToUIntArray(block)
            For j As Integer = 0 To 127
                Dim temp As UInt32 = uddata(j)
                If CBool(j And &H1) Then LP5 = LP5 Xor temp Else LP4 = LP4 Xor temp
                If CBool(j And &H2) Then LP7 = LP7 Xor temp Else LP6 = LP6 Xor temp
                If CBool(j And &H4) Then LP9 = LP9 Xor temp Else LP8 = LP8 Xor temp
                If CBool(j And &H8) Then LP11 = LP11 Xor temp Else LP10 = LP10 Xor temp
                If CBool(j And &H10) Then LP13 = LP13 Xor temp Else LP12 = LP12 Xor temp
                If CBool(j And &H20) Then LP15 = LP15 Xor temp Else LP14 = LP14 Xor temp
                If CBool(j And &H40) Then LP17 = LP17 Xor temp Else LP16 = LP16 Xor temp
            Next
            Dim reg32 As UInt32 = (LP15 Xor LP14)
            Dim byte_reg As Byte = CByte((reg32) And &HFF)
            byte_reg = byte_reg Xor CByte((reg32 >> 8) And &HFF)
            byte_reg = byte_reg Xor CByte((reg32 >> 16) And &HFF)
            byte_reg = byte_reg Xor CByte((reg32 >> 24) And &HFF)
            byte_reg = byte_parity_table(byte_reg)
            word_reg = (CUShort(LP16 >> 16)) Xor (CUShort(LP16 And &HFFFF))
            LP16 = CByte(CByte(word_reg And &HFF) Xor CByte(word_reg >> 8))
            word_reg = CUShort(LP17 >> 16) Xor CUShort(LP17 And &HFFFF)
            LP17 = CByte(CByte(word_reg And &HFF) Xor CByte(word_reg >> 8))
            ecc_code(2) = ((byte_reg And CByte(&HFE)) << 1) Or byte_parity_table(CByte(LP16 And &HFF) And CByte(&H1)) Or (byte_parity_table(CByte(LP17 And &HFF) And CByte(&H1)) << 1)
            LP0 = CByte((reg32 Xor (reg32 >> 16)) And 255)
            LP1 = CByte(((reg32 >> 8) Xor (reg32 >> 24)) And 255)
            LP2 = CByte((reg32 Xor (reg32 >> 8)) And 255)
            LP3 = CByte(((reg32 >> 16) Xor (reg32 >> 24)) And 255)
            word_reg = CUShort(LP4 >> 16) Xor CUShort(LP4 And &HFFFF)
            LP4 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP5 >> 16) Xor CUShort(LP5 And &HFFFF)
            LP5 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP6 >> 16) Xor CUShort(LP6 And &HFFFF)
            LP6 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP7 >> 16) Xor CUShort(LP7 And &HFFFF)
            LP7 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP8 >> 16) Xor CUShort(LP8 And &HFFFF)
            LP8 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP9 >> 16) Xor CUShort(LP9 And &HFFFF)
            LP9 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP10 >> 16) Xor CUShort(LP10 And &HFFFF)
            LP10 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP11 >> 16) Xor CUShort(LP11 And &HFFFF)
            LP11 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP12 >> 16) Xor CUShort(LP12 And &HFFFF)
            LP12 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP13 >> 16) Xor CUShort(LP13 And &HFFFF)
            LP13 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP14 >> 16) Xor CUShort(LP14 And &HFFFF)
            LP14 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))
            word_reg = CUShort(LP15 >> 16) Xor CUShort(LP15 And &HFFFF)
            LP15 = CByte(CByte(word_reg And 255) Xor (CByte(word_reg >> 8)))

            ecc_code(0) = (byte_parity_table(CByte(LP0 And 255)) And CByte(&H1)) Or
                ((byte_parity_table(CByte(LP1 And 255)) And CByte(&H1)) << 1) Or
                ((byte_parity_table(CByte(LP2 And 255)) And CByte(&H1)) << 2) Or
                ((byte_parity_table(CByte(LP3 And 255)) And CByte(&H1)) << 3) Or
                ((byte_parity_table(CByte(LP4 And 255)) And CByte(&H1)) << 4) Or
                ((byte_parity_table(CByte(LP5 And 255)) And CByte(&H1)) << 5) Or
                ((byte_parity_table(CByte(LP6 And 255)) And CByte(&H1)) << 6) Or
                ((byte_parity_table(CByte(LP7 And 255)) And CByte(&H1)) << 7)

            ecc_code(1) = ((byte_parity_table(CByte(LP8 And 255)) And CByte(&H1))) Or
                (byte_parity_table(CByte(LP9 And 255)) And CByte(&H1)) << 1 Or
                (byte_parity_table(CByte(LP10 And 255)) And CByte(&H1)) << 2 Or
                (byte_parity_table(CByte(LP11 And 255)) And CByte(&H1)) << 3 Or
                (byte_parity_table(CByte(LP12 And 255)) And CByte(&H1)) << 4 Or
                (byte_parity_table(CByte(LP13 And 255)) And CByte(&H1)) << 5 Or
                (byte_parity_table(CByte(LP14 And 255)) And CByte(&H1)) << 6 Or
                (byte_parity_table(CByte(LP15 And 255)) And CByte(&H1)) << 7

            Return ecc_code
        End Function

        Public Function ProcessECC(ByRef block() As Byte, ByRef stored_ecc() As Byte) As ECC_DECODE_RESULT Implements EccEngine.ProcessECC
            Dim new_ecc() As Byte = GenerateECC(block)
            Dim ecc_xor(2) As Byte
            ecc_xor(0) = new_ecc(0) Xor stored_ecc(0)
            ecc_xor(1) = new_ecc(1) Xor stored_ecc(1)
            ecc_xor(2) = new_ecc(2) Xor stored_ecc(2)
            If ((ecc_xor(0) Or ecc_xor(1) Or ecc_xor(2)) = 0) Then
                Return ECC_DECODE_RESULT.NoErrors
            Else
                Dim bit_count As Integer = BitCount(ecc_xor) 'Counts the bit number
                If (bit_count = 12) Then
                    Dim bit_addr As Byte = ((ecc_xor(2) >> 3) And CByte(1)) Or ((ecc_xor(2) >> 4) And CByte(2)) Or ((ecc_xor(2) >> 5) And CByte(4))
                    Dim byte_addr As UInt16 =
                    (ecc_xor(0) >> 1) And &H1US Or
                    (ecc_xor(0) >> 2) And &H2US Or
                    (ecc_xor(0) >> 3) And &H4US Or
                    (ecc_xor(0) >> 4) And &H8US Or
                    (ecc_xor(1) << 3) And &H10US Or
                    (ecc_xor(1) << 2) And &H20US Or
                    (ecc_xor(1) << 1) And &H40US Or
                    (ecc_xor(1) << 0) And &H80US Or
                    (ecc_xor(2) << 7) And &H100US
                    block(byte_addr - 1) = block(byte_addr - 1) Xor (CByte(1) << bit_addr)
                    Return ECC_DECODE_RESULT.Correctable
                ElseIf (bit_count = 1) Then
                    stored_ecc = new_ecc
                    Return ECC_DECODE_RESULT.EccError
                Else
                    Return ECC_DECODE_RESULT.Uncorractable
                End If
            End If
        End Function
        'Returns the number of bytes this ECC requires per block of data
        Public Function ECC_Size() As Integer Implements EccEngine.ECC_Size
            Return 3
        End Function

        Public Sub Close() Implements EccEngine.Close
            'Not needed
        End Sub

    End Class

    Public Class RS_ECC : Implements EccEngine
        Private RScontrol As ECC.ReedSolomon
        Private BlockSize As Integer
        Public ReadOnly Property PARITY_BITS As Integer 'RS_CONST_T (Number of symbols to correct)
        Public ReadOnly Property SYM_WIDTH As Integer 'RS_SYM_W (Number of bits per symbol), 8, 9 or 10

        Sub New(block_size As Integer, parity_bits As Integer, sym_width As Integer)
            Me.PARITY_BITS = parity_bits
            Me.SYM_WIDTH = sym_width
            BlockSize = block_size
            RScontrol = New ECC.ReedSolomon(Me.SYM_WIDTH, GetPolynomial(Me.SYM_WIDTH), 0, 2, Me.PARITY_BITS * 2)
        End Sub

        Public Function GenerateECC(block() As Byte) As Byte() Implements EccEngine.GenerateECC
            Dim Symbols() As Integer = bytes_to_symbols(block, Me.SYM_WIDTH)
            Dim result() As Integer = RScontrol.Encode(Symbols)
            Dim ecc_data() As Byte = symbols_to_bytes(result, Me.SYM_WIDTH)
            Return ecc_data
        End Function

        Public Function ProcessECC(ByRef block() As Byte, ByRef stored_ecc() As Byte) As ECC_DECODE_RESULT Implements EccEngine.ProcessECC
            Dim Symbols() As Integer = bytes_to_symbols(block, Me.SYM_WIDTH)
            Dim EccData() As Integer = bytes_to_symbols(stored_ecc, Me.SYM_WIDTH)
            Dim cmp_result As ECC.CompareResult = RScontrol.Decode(Symbols, EccData)
            Select Case cmp_result
                Case ECC.CompareResult.NoError
                    Return ECC_DECODE_RESULT.NoErrors
                Case ECC.CompareResult.EccError
                    Dim org_size As Integer = stored_ecc.Length
                    stored_ecc = symbols_to_bytes(EccData, Me.SYM_WIDTH)
                    ReDim Preserve stored_ecc(org_size - 1)
                    Return ECC_DECODE_RESULT.EccError
                Case ECC.CompareResult.Correctable
                    Dim org_size As Integer = block.Length
                    block = symbols_to_bytes(Symbols, Me.SYM_WIDTH)
                    ReDim Preserve block(org_size - 1)
                    Return ECC_DECODE_RESULT.Correctable
                Case ECC.CompareResult.BothErrors
                    Dim org_data_size As Integer = block.Length
                    block = symbols_to_bytes(Symbols, Me.SYM_WIDTH)
                    ReDim Preserve block(org_data_size - 1)
                    Dim org_ecc_size As Integer = stored_ecc.Length
                    stored_ecc = symbols_to_bytes(EccData, Me.SYM_WIDTH)
                    ReDim Preserve stored_ecc(org_ecc_size - 1)
                    Return ECC_DECODE_RESULT.DataEccError
                Case Else
                    Return ECC_DECODE_RESULT.Uncorractable
            End Select
        End Function

        Private Shared Function GetPolynomial(sym_size As Integer) As Integer
            Select Case sym_size
                Case 0 'dont care
                Case 1 'dont care
                Case 2 '2-nd: poly = x^2 + x + 1
                    Return &H7
                Case 3 '3-rd: poly = x^3 + x + 1
                    Return &HB
                Case 4  '4-th: poly = x^4 + x + 1
                    Return &H13
                Case 5 '5-th: poly = x^5 + x^2 + 1
                    Return &H25
                Case 6 '6-th: poly = x^6 + x + 1
                    Return &H43
                Case 7 '7-th: poly = x^7 + x^3 + 1
                    Return &H89
                Case 8 '8-th: poly = x^8 + x^4 + x^3 + x^2 + 1
                    Return &H11D
                Case 9 '9-th: poly = x^9 + x^4 + 1 
                    Return &H211
                Case 10  '10-th: poly = x^10 + x^3 + 1
                    Return &H409
                Case 11 '11-th: poly = x^11 + x^2 + 1
                    Return &H805
                Case 12 '12-th: poly = x^12 + x^6 + x^4 + x + 1
                    Return &H1053
                Case 13 '13-th: poly = x^13 + x^4 + x^3 + x + 1
                    Return &H201B
                Case 14 '14-th: poly = x^14 + x^10 + x^6 + x + 1
                    Return &H4443
                Case 15 '15-th: poly = x^15 + x + 1
                    Return &H8003
                Case 16 '16-th: poly = x^16 + x^12 + x^3 + x + 1
                    Return &H1100B
            End Select
            Return -1
        End Function
        'Converts data stored in 32-bit to byte by the specified symbol width
        Private Shared Function symbols_to_bytes(data_in() As Integer, sym_width As Integer) As Byte()
            Dim ecc_size As Integer = CInt(Math.Ceiling((data_in.Length * sym_width) / 8))
            Dim data_out(ecc_size - 1) As Byte 'This needs to be 0x00
            Dim counter As Integer = 0
            Dim spare_data As Byte = 0
            Dim bits_left As Integer = 0
            For Each item As Integer In data_in
                Do
                    Dim next_bits As Integer = (8 - bits_left)
                    bits_left = (sym_width - next_bits)
                    Dim byte_to_add As Byte = CByte((spare_data << next_bits) Or ((item >> bits_left) And ((1 << next_bits) - 1)))
                    data_out(counter) = byte_to_add : counter += 1
                    Dim x As Integer = Math.Min(bits_left, 8)
                    Dim offset As Integer = (bits_left - x)
                    spare_data = CByte((item And ((1 << x) - 1) << offset) >> offset)
                    If x = 8 Then
                        data_out(counter) = spare_data : counter += 1
                        bits_left -= 8
                        spare_data = CByte(item And ((1 << offset) - 1))
                    End If
                Loop While bits_left > 8
            Next
            If (bits_left > 0) Then
                Dim next_bits As Integer = (8 - bits_left)
                Dim byte_to_add As Byte = (spare_data << next_bits)
                data_out(counter) = byte_to_add  'Last byte
            End If
            Return data_out
        End Function

        Private Shared Function bytes_to_symbols(data_in() As Byte, sym_width As Integer) As Integer()
            Dim sym_count As Integer = CInt(Math.Ceiling((data_in.Length * 8) / sym_width))
            Dim data_out(sym_count - 1) As Integer
            Dim counter As Integer = 0
            Dim int_data As Integer
            Dim bit_offset As Integer = 0
            For i = 0 To data_in.Length - 1
                Dim data_in_bitcount As Integer = 8
                Do
                    Dim bits_left As Integer = (sym_width - bit_offset) 'number of bits our int_data needed
                    Dim sel_count As Integer = Math.Min(bits_left, data_in_bitcount) 'number of bits we can pull from the current byte
                    Dim target_offset As Integer = sym_width - (bit_offset + sel_count)
                    data_in_bitcount -= sel_count
                    Dim src_offset As Integer = data_in_bitcount
                    Dim bit_mask As Byte = CByte(((1 << sel_count) - 1) << src_offset)
                    Dim data_selected As Integer = (data_in(i) And bit_mask) >> src_offset
                    int_data = int_data Or (data_selected << target_offset)
                    bit_offset += sel_count
                    If bit_offset = sym_width Then
                        data_out(counter) = int_data : counter += 1
                        bit_offset = 0
                        int_data = 0
                    End If
                Loop While data_in_bitcount > 0
            Next
            If (bit_offset > 0) Then
                data_out(counter) = int_data
            End If
            Return data_out
        End Function

        Public Shared Function PredictSize(block_size As Integer, bit_error As Integer, sym_width As Integer) As Integer
            Using rs As New ECC.ReedSolomon(sym_width, GetPolynomial(sym_width), 0, 2, bit_error * 2)
                Dim block(block_size - 1) As Byte
                For i = 0 To block.Length - 1
                    block(i) = CByte(Utilities.RandomNumber(0, 255))
                Next
                Dim Symbols() As Integer = bytes_to_symbols(block, sym_width)
                Dim result() As Integer = rs.Encode(Symbols)
                Dim ecc_data() As Byte = symbols_to_bytes(result, sym_width)
                Return ecc_data.Length
            End Using
        End Function

        Public Function ECC_Size() As Integer Implements EccEngine.ECC_Size
            Return PredictSize(Me.BlockSize, Me.PARITY_BITS, Me.SYM_WIDTH)
        End Function

        Public Sub Close() Implements EccEngine.Close
            RScontrol.Dispose()
            RScontrol = Nothing
        End Sub

    End Class

    Public Class BinaryBCH : Implements EccEngine
        'Galois field order, should be in the range 5-15
        '5=1 byte ([1..5] bits)
        '6=4 bytes ([1..9] bits)
        '7=8 bytes ([1..17] bits)
        '8=16 bytes ([1..30] bits)
        '9=32 bytes ([1..55] bits)
        '10=64 bytes ([1..101] bits)
        '11=128 bytes ([1..185] bits)
        '12=256 bytes ([1..340] bits)
        '13=512 bytes ([1..629] bits)
        '14=1024 bytes ([1..1169] bits)
        '15=2048 bytes ([1..2183] bits)
        Private ReadOnly Property BCH_CONST_M As Integer
        Private ReadOnly Property BCH_CONST_T As Integer

        Private BchControl As ECC.BCH

        'Call this with the size of data to be corrected and how many bits at max can be corrected
        Sub New(block_size As Integer, parity_bits As Integer)
            Me.BCH_CONST_M = Get_CONST_M(block_size)
            Me.BCH_CONST_T = parity_bits
            Dim c = Get_MaxBitCorrection(block_size)
            BchControl = New ECC.BCH(Me.BCH_CONST_M, Me.BCH_CONST_T)
            If (c = -1 OrElse parity_bits > c) Then
                Throw New Exception("Parity bits must be between 1 and " & c.ToString())
            End If
        End Sub

        Public Function GenerateECC(block() As Byte) As Byte() Implements EccEngine.GenerateECC
            Return BchControl.Encode(block)
        End Function

        Public Function ProcessECC(ByRef block() As Byte, ByRef stored_ecc() As Byte) As ECC_DECODE_RESULT Implements EccEngine.ProcessECC
            Dim cmp_result As ECC.CompareResult = BchControl.Decode(block, stored_ecc)
            Select Case cmp_result
                Case ECC.CompareResult.NoError
                    Return ECC_DECODE_RESULT.NoErrors
                Case ECC.CompareResult.EccError
                    Return ECC_DECODE_RESULT.EccError
                Case ECC.CompareResult.Correctable
                    Return ECC_DECODE_RESULT.Correctable
                Case ECC.CompareResult.BothErrors 'The ECC and the DATA had errors
                    Return ECC_DECODE_RESULT.DataEccError
                Case ECC.CompareResult.Uncorractable
                    Return ECC_DECODE_RESULT.Uncorractable
            End Select
            Return ECC_DECODE_RESULT.Unknown
        End Function

        Private Shared Function Get_CONST_M(size_of_block As Integer) As Integer
            If size_of_block = 1 Then
                Return 5
            ElseIf size_of_block <= 4 Then
                Return 6
            ElseIf size_of_block <= 8 Then
                Return 7
            ElseIf size_of_block <= 16 Then
                Return 8
            ElseIf size_of_block <= 32 Then
                Return 9
            ElseIf size_of_block <= 64 Then
                Return 10
            ElseIf size_of_block <= 128 Then
                Return 11
            ElseIf size_of_block <= 256 Then
                Return 12
            ElseIf size_of_block <= 512 Then
                Return 13
            ElseIf size_of_block <= 1024 Then
                Return 14
            ElseIf size_of_block <= 2048 Then
                Return 15
            Else
                Throw New Exception("Block size must be between 1 and 2048 bytes")
                Return -1
            End If
        End Function

        Public Shared Function Get_MaxBitCorrection(size_of_block As Integer) As Integer
            If size_of_block = 1 Then
                Return 5
            ElseIf size_of_block <= 4 Then
                Return 9
            ElseIf size_of_block <= 8 Then
                Return 17
            ElseIf size_of_block <= 16 Then
                Return 30
            ElseIf size_of_block <= 32 Then
                Return 55
            ElseIf size_of_block <= 64 Then
                Return 101
            ElseIf size_of_block <= 128 Then
                Return 185
            ElseIf size_of_block <= 256 Then
                Return 340
            ElseIf size_of_block <= 512 Then
                Return 629
            ElseIf size_of_block <= 1024 Then
                Return 1169
            ElseIf size_of_block <= 2048 Then
                Return 2183
            Else
                Return -1
            End If
        End Function

        Public Shared Function PredictSize(block_size As UInt16, bit_error As UInt16) As Integer
            Dim m = Get_CONST_M(block_size)
            Return CInt(Math.Ceiling((m * bit_error) / 8))
        End Function

        Public Function ECC_Size() As Integer Implements EccEngine.ECC_Size
            Return CInt(Math.Ceiling((BCH_CONST_M * BCH_CONST_T) / 8))
        End Function

        Public Sub Close() Implements EccEngine.Close
            BchControl.Dispose()
            BchControl = Nothing
        End Sub

    End Class

    Public Class DharaBCH : Implements EccEngine
        Private BlockSize As Integer
        Private ParityBits As Integer
        Private DharaControl As ECC.BCH_Dhara

        'parity_bits = 1..4
        Sub New(block_size As Integer, parity_bits As Integer)
            Me.BlockSize = block_size
            Me.ParityBits = parity_bits
            DharaControl = New ECC.BCH_Dhara(parity_bits)
        End Sub

        Public Function GenerateECC(block() As Byte) As Byte() Implements EccEngine.GenerateECC
            Dim ecc = DharaControl.Encode(block)
            Return ecc
        End Function

        Public Function ProcessECC(ByRef block() As Byte, ByRef stored_ecc() As Byte) As ECC_DECODE_RESULT Implements EccEngine.ProcessECC
            Dim cmp_result = DharaControl.Decode(block, stored_ecc)
            Select Case cmp_result
                Case ECC.CompareResult.NoError
                    Return ECC_DECODE_RESULT.NoErrors
                Case ECC.CompareResult.Correctable
                    Return ECC_DECODE_RESULT.Correctable
                Case ECC.CompareResult.Uncorractable
                    Return ECC_DECODE_RESULT.Uncorractable
            End Select
            Return ECC_DECODE_RESULT.Unknown
        End Function

        Public Function GetEccSize() As Integer Implements EccEngine.ECC_Size
            Select Case Me.ParityBits
                Case 1
                    Return 2
                Case 2
                    Return 4
                Case 3
                    Return 5
                Case 4
                    Return 7
                Case Else
                    Return 0
            End Select
        End Function

        Public Shared Function PredictSize(block_size As UInt16, bit_error As UInt16) As Integer
            Dim n As New DharaBCH(block_size, bit_error)
            Dim block(block_size - 1) As Byte
            For i = 0 To block.Length - 1
                block(i) = CByte(Utilities.RandomNumber(0, 255))
            Next
            Dim d() = n.GenerateECC(block)
            Return d.Length
        End Function

        Public Sub Close() Implements EccEngine.Close
            'Not needed
        End Sub

    End Class

    Public Class ECC_Engine : Implements ECC_Engine_Interface
        Private ecc_mode As ecc_algorithum
        Private ecc_engine As EccEngine = Nothing
        Public LastResult As ECC_DECODE_RESULT = ECC_DECODE_RESULT.NoErrors 'Contains the last decoded result
        Public LastPageIndex As Integer 'Contains the page index that had an error (-1 means no error was recorded)

        Public ReadOnly Property Chunk_Size As UInt16 'Number of bytes to process for each ECC
        Public ReadOnly Property RS_SYM_WIDTH As Integer = 9 'Only used by RS
        Public Property REVERSE_ARRAY As Boolean = False 'RS option allows to reverse the input byte array
        Public Property InitialOffset As UInt16
        Public Property Spacing As UInt16

        Sub New(mode As ecc_algorithum, block_size As UShort, Optional bits_correction As Integer = 1, Optional rs_sym_width As Integer = 0)
            Me.ecc_mode = mode
            Me.RS_SYM_WIDTH = rs_sym_width
            Me.Chunk_Size = block_size
            Select Case ecc_mode
                Case ecc_algorithum.hamming 'Hamming only supports 1-bit ECC correction
                    ecc_engine = New Hamming()
                Case ecc_algorithum.reedsolomon
                    ecc_engine = New RS_ECC(block_size, bits_correction, rs_sym_width)
                Case ecc_algorithum.bch
                    ecc_engine = New BinaryBCH(block_size, bits_correction)
                Case ecc_algorithum.dhara
                    ecc_engine = New DharaBCH(block_size, bits_correction)
            End Select
        End Sub

        Public Function GenerateECC(data_in() As Byte) As Byte()
            If Me.REVERSE_ARRAY Then Array.Reverse(data_in)
            Return ecc_engine.GenerateECC(data_in)
        End Function

        Public Function ProcessECC(ByRef data_in() As Byte, ByRef ecc_codes() As Byte) As ECC_DECODE_RESULT
            Return ecc_engine.ProcessECC(data_in, ecc_codes)
        End Function

        'Processes blocks and returns the last decoded result
        Public Function ReadData(ByRef flash_block() As Byte, page_size As UInt16, ByRef flash_spare_area() As Byte, spare_size As UInt16) As ECC_DECODE_RESULT Implements ECC_Engine_Interface.ReadData
            Try
                If (Not flash_block.Length Mod page_size = 0) Then Return ECC_DECODE_RESULT.Unknown
                Dim page_count As Integer = CInt(flash_block.Length / page_size)
                For page_index = 0 To page_count - 1
                    Dim page_data(page_size - 1) As Byte
                    Dim oob_data(spare_size - 1) As Byte
                    Array.Copy(flash_block, page_index * page_size, page_data, 0, page_data.Length)
                    Array.Copy(flash_spare_area, page_index * spare_size, oob_data, 0, oob_data.Length)
                    LastResult = ReadData_CorrectECC(page_data, oob_data)
                    If LastResult = ECC_DECODE_RESULT.Correctable Then
                        Array.Copy(page_data, 0, flash_block, page_index * page_size, page_data.Length) 'We corrected data, lets put it back into the main buffer
                    ElseIf LastResult = ECC_DECODE_RESULT.Uncorractable Then
                        Me.LastPageIndex = page_index
                        Return LastResult
                    End If
                Next
                Me.LastPageIndex = -1
                LastResult = ECC_DECODE_RESULT.NoErrors 'We corrected any errors and the buffer now contains no errors
            Catch ex As Exception
                LastResult = ECC_DECODE_RESULT.Unknown
            End Try
            Return LastResult
        End Function

        Private Function ReadData_CorrectECC(ByRef flash_page() As Byte, ByRef flash_spare() As Byte) As ECC_DECODE_RESULT
            If (flash_page Is Nothing OrElse (flash_page.Length < Me.Chunk_Size)) Then Return ECC_DECODE_RESULT.Unknown
            Dim ecc_byte_size As Integer = ecc_engine.ECC_Size()
            Dim ecc_ptr As Integer = Me.InitialOffset
            Dim ecc_count = CInt(Math.Ceiling(flash_page.Length / Me.Chunk_Size))
            Dim bytes_left = flash_page.Length
            Dim ecc_result = ECC_DECODE_RESULT.NoErrors
            For i = 0 To ecc_count - 1
                Dim base_addr = (i * Me.Chunk_Size)
                Dim chunk_size = Math.Min(Me.Chunk_Size, bytes_left)
                Dim chunk(chunk_size - 1) As Byte
                Array.Copy(flash_page, base_addr, chunk, 0, chunk.Length)
                If Me.REVERSE_ARRAY Then Array.Reverse(chunk)
                Dim ecc_data(ecc_byte_size - 1) As Byte
                Array.Copy(flash_spare, ecc_ptr, ecc_data, 0, ecc_data.Length)
                If Not Utilities.IsByteArrayFilled(ecc_data, &HFF) Then 'We only want to process ECC data that is not blank
                    Dim ecc_call_result As ECC_DECODE_RESULT = ecc_engine.ProcessECC(chunk, ecc_data)
                    If ecc_call_result = ECC_DECODE_RESULT.Correctable Or ecc_call_result = ECC_DECODE_RESULT.DataEccError Then 'The DATA has an error and has been corrected 
                        Array.Copy(chunk, 0, flash_page, base_addr, chunk.Length)
                        ecc_result = ECC_DECODE_RESULT.Correctable 'To let our calling function know that we need to copy the good data
                    ElseIf ecc_call_result = ECC_DECODE_RESULT.EccError Then 'The ECC has an error but we can ignore (used by Hamming)
                    ElseIf ecc_call_result = ECC_DECODE_RESULT.Uncorractable Then
                        Return ECC_DECODE_RESULT.Uncorractable
                    ElseIf ecc_call_result = ECC_DECODE_RESULT.NoErrors Then 'Nothing to do
                    End If
                End If
                ecc_ptr += ecc_byte_size + Me.Spacing
                bytes_left -= chunk_size
            Next
            Return ecc_result
        End Function

        Public Sub WriteData(ByRef flash_block() As Byte, page_size As UInt16, ByRef flash_spare_area() As Byte, spare_size As UInt16) Implements ECC_Engine_Interface.WriteData
            Try
                If (Not flash_block.Length Mod page_size = 0) Then Exit Sub
                Dim page_count As Integer = CInt(flash_block.Length / page_size)
                For page_index = 0 To page_count - 1
                    Dim page_data(page_size - 1) As Byte
                    Dim oob_data(spare_size - 1) As Byte
                    Array.Copy(flash_block, page_index * page_size, page_data, 0, page_data.Length)
                    Array.Copy(flash_spare_area, page_index * spare_size, oob_data, 0, oob_data.Length)
                    WriteData_GenerateECC(page_data, oob_data) 'Generate the ECC and store it correctly within the page oob area
                    Array.Copy(oob_data, 0, flash_spare_area, page_index * spare_size, oob_data.Length) 'Copy the data with the new ECC back to the spare area
                Next
            Catch ex As Exception
            End Try
        End Sub

        Private Sub WriteData_GenerateECC(flash_page() As Byte, ByRef flash_spare() As Byte)
            Dim ecc_byte_size As Integer = ecc_engine.ECC_Size()
            Dim ecc_ptr As Integer = Me.InitialOffset
            Dim ecc_count = CInt(Math.Ceiling(flash_page.Length / Me.Chunk_Size))
            Dim bytes_left = flash_page.Length
            For i = 0 To ecc_count - 1
                Dim base_addr = (i * Me.Chunk_Size)
                Dim block_size = Math.Min(Me.Chunk_Size, bytes_left)
                Dim block(block_size - 1) As Byte
                Array.Copy(flash_page, base_addr, block, 0, block.Length)
                If Me.REVERSE_ARRAY Then Array.Reverse(block)
                Dim ecc_data() As Byte = ecc_engine.GenerateECC(block)
                Array.Copy(ecc_data, 0, flash_spare, ecc_ptr, ecc_data.Length) 'Copies the new ECC codes to the destination ecc buffer
                ecc_ptr += ecc_byte_size + Me.Spacing
                bytes_left -= block_size
            Next
        End Sub

        Public Sub Close()
            ecc_engine.Close()
        End Sub

    End Class

End Namespace